from django import forms
from django.contrib.auth import authenticate, forms as auth_forms
from django.utils import timezone
from django.utils.translation import gettext_lazy as _


def promo_choices():
    return [(r, r) for r in range(2000, timezone.now().year + 1)]


class OldCASAuthForm(forms.Form):
    """Adapts Django's AuthenticationForm to allow for OldCAS login."""

    cas_login = auth_forms.UsernameField(
        label=_("Ancien login clipper"), max_length=1023
    )
    password = forms.CharField(
        label=_("Mot de passe"),
        strip=False,
        widget=forms.PasswordInput(attrs={"autocomplete": "current-password"}),
    )
    entrance_year = forms.TypedChoiceField(
        label=_("Promotion"), choices=promo_choices, coerce=int
    )

    def __init__(self, request=None, *args, **kwargs):
        self.request = request
        self.user_cache = None
        super().__init__(*args, **kwargs)

    def clean(self):
        cas_login = self.cleaned_data.get("cas_login")
        password = self.cleaned_data.get("password")
        entrance_year = self.cleaned_data.get("entrance_year")

        if cas_login is not None and password:
            self.user_cache = authenticate(
                self.request,
                cas_login=cas_login,
                password=password,
                entrance_year=entrance_year,
            )
            if self.user_cache is None:
                raise self.get_invalid_login_error()

        return self.cleaned_data

    def get_user(self):
        # Necessary API for LoginView
        return self.user_cache

    def get_invalid_login_error(self):
        return forms.ValidationError(
            _(
                "Aucun utilisateur n'existe avec ce clipper, cette promo et/ou ce mot "
                "de passe. Veuillez vérifier votre saisie. Attention, tous les champs "
                "sont sensibles à la casse !"
            ),
            code="invalid_login",
        )



class FakeCASClient:
    """Fake CAS Client that mimics CAS interactions.

    By default, always log in the same user 'johndoe' without checking the ticket.
    """

    def __init__(self, cas_login="johndoe", entrance_year=2019):
        self.cas_login = cas_login
        self.entrance_year = entrance_year

    def verify_ticket(self, ticket):
        """Dummy client that always log in the same user."""

        attributes = {
            "homeDirectory": "/users/{}/info/{}".format(
                self.entrance_year % 100, self.cas_login
            )
        }
        return self.cas_login, attributes, None


class FakeLDAPObject:
    """Fake object to be used in place of the result of `ldap.initialize`.

    Always return the same user information (configured at class initialization).
    """

    def __init__(self, cas_login: str, entrance_year: int):
        self.cas_login = cas_login
        self.entrance_year = entrance_year

    def search_s(self, base, scope, request, *args):
        if request != "(uid={})".format(self.cas_login):
            raise ValueError("I don't know how to answer this request!")

        home_dir = "/users/{}/info/{}".format(self.entrance_year % 100, self.cas_login)
        dn = "whatever"
        attrs = {
            "uid": [self.cas_login.encode("utf-8")],
            "cn": ["{}'s long name".format(self.cas_login).encode("utf-8")],
            "homeDirectory": [home_dir.encode("utf-8")],
        }
        return [(dn, attrs)]



from urllib.parse import urlparse, urlunparse

from django.conf import settings
from django.contrib import auth
from django.contrib.auth import views as auth_views
from django.contrib.messages.views import SuccessMessageMixin
from django.core.exceptions import PermissionDenied
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.views.generic import TemplateView, View

from authens import conf as default_conf
from authens.forms import OldCASAuthForm
from authens.utils import get_cas_client


class NextPageMixin:
    def get_next_url(self):
        """Decide where to go after a successful login.

        Look for (in order):
        - a `next` GET parameter;
        - a `CASNEXT` session variable;
        - the `LOGIN_REDIRECT_URL` django setting.
        """
        request = self.request
        next_url = request.GET.get("next")
        if next_url is None and "CASNEXT" in request.session:
            next_url = request.session["CASNEXT"]
            del request.session["CASNEXT"]
        if next_url is None:
            next_url = settings.LOGIN_REDIRECT_URL
        return str(next_url)


class LoginSwitchView(NextPageMixin, TemplateView):
    """Simple page letting the user choose between password and CAS authentication."""

    template_name = "authens/login_switch.html"
    http_method_names = ["get"]

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            return redirect(self.get_next_url())
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)

        ctx.update(
            {
                "next": self.get_next_url(),
                "oldcas": getattr(
                    settings, "AUTHENS_USE_OLDCAS", default_conf.AUTHENS_USE_OLDCAS
                ),
                "password": getattr(
                    settings, "AUTHENS_USE_PASSWORD", default_conf.AUTHENS_USE_PASSWORD
                ),
            }
        )
        return ctx


class CASLoginView(NextPageMixin, View):
    """CAS authentication view.

    Implement the CAS authentication scheme:

    1. We first redirect the user to the student CAS.
    2. The user comes back with a ticket, we validate it to make sure the user is legit
       (validation is delegated to the ENSCASBackend).
    3. We redirect the user to the next page.
    """

    http_method_names = ["get"]

    def get(self, request, *args, **kwargs):
        ticket = request.GET.get("ticket")

        if not ticket:
            request.session["CASNEXT"] = self.get_next_url()
            cas_client = get_cas_client(request)
            return redirect(cas_client.get_login_url())

        user = auth.authenticate(request, ticket=ticket)
        if user is None:
            raise PermissionDenied(_("Connection échouée !"))
        auth.login(request, user)
        return redirect(self.get_next_url())


class PasswordLoginView(auth_views.LoginView):
    template_name = "authens/pwd_login.html"


class PasswordResetView(SuccessMessageMixin, auth_views.PasswordResetView):
    template_name = "authens/pwd_reset.html"
    email_template_name = "authens/pwd_reset_email.txt"
    subject_template_name = "authens/pwd_reset_subject.txt"
    success_url = reverse_lazy("authens:login")

    success_message = _(
        "Si un compte avec cet email existe, un email de réinitialisation vient de lui être envoyé !"
    )


class PasswordResetConfirmView(
    SuccessMessageMixin, auth_views.PasswordResetConfirmView
):
    template_name = "authens/pwd_reset_confirm.html"
    success_url = reverse_lazy("authens:login")

    success_message = _("Mot de passe modifié avec succès !")


class OldCASLoginView(auth_views.LoginView):
    template_name = "authens/oldcas_login.html"
    authentication_form = OldCASAuthForm

    def get_initial(self):
        return {"entrance_year": timezone.now().year - 5}


class LogoutView(auth_views.LogoutView):
    """Logout view of AuthENS.

    Tell Django to log the user out, then redirect to the CAS logout page if the user
    logged in via CAS.
    """

    def setup(self, request):
        super().setup(request)
        if "CASCONNECTED" in request.session:
            del request.session["CASCONNECTED"]
            self.cas_connected = True
        else:
            self.cas_connected = False

    def get_next_page(self):
        next_page = super().get_next_page()
        if self.cas_connected:
            cas_client = get_cas_client(self.request)

            # If the next_url is local (no hostname), make it absolute so that the user
            # is correctly redirected from CAS.
            if next_page is not None and not urlparse(next_page).netloc:
                request = self.request
                next_page = urlunparse(
                    (request.scheme, request.get_host(), next_page, "", "", "")
                )

            next_page = cas_client.get_logout_url(redirect_url=next_page)
        return next_page



from unittest import mock

from django.contrib.auth import get_user_model
from django.test import TestCase

from authens.models import CASAccount, OldCASAccount
from authens.shortcuts import register_cas_account
from authens.tests.cas_utils import FakeLDAPObject

User = get_user_model()


class TestRegisterCasAccount(TestCase):
    @mock.patch("authens.shortcuts.ldap.initialize")
    def test_register(self, mock_ldap_obj):
        mock_ldap_obj.return_value = FakeLDAPObject("johndoe", 2019)

        user = User.objects.create_user(username="whatever")
        self.assertFalse(hasattr(user, "cas_account"))

        register_cas_account(user, cas_login="johndoe")
        user.refresh_from_db()
        self.assertTrue(hasattr(user, "cas_account"))
        self.assertEqual(user.cas_account.cas_login, "johndoe")
        self.assertEqual(user.cas_account.entrance_year, 2019)

    def test_cant_register_twice(self):
        john = User.objects.create_user(username="whatever")
        CASAccount.objects.create(user=john, cas_login="johndoe", entrance_year=2019)

        janis = User.objects.create_user(username="janisjoplin")

        # John cannot have two CAS accounts
        with self.assertRaises(ValueError):
            register_cas_account(john, cas_login="joplin")

        # Janis cannot steal John's account
        with self.assertRaises(ValueError):
            register_cas_account(janis, cas_login="johndoe")

        self.assertEqual(CASAccount.objects.count(), 1)

    def test_cant_register_old_account(self):
        user = User.objects.create_user(username="whatever")
        OldCASAccount.objects.create(user=user, cas_login="toto", entrance_year=2012)

        with self.assertRaises(ValueError):
            register_cas_account(user, cas_login="toto")

        self.assertFalse(CASAccount.objects.exists())



from unittest.mock import patch
from urllib.parse import quote

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sessions.models import Session
from django.test import Client, TestCase
from django.urls import reverse

from authens.models import CASAccount, OldCASAccount
from authens.tests.cas_utils import FakeCASClient

UserModel = get_user_model()


class TestLoginViews(TestCase):
    def test_cas_login_redirect(self):
        url = reverse("authens:login.cas")
        abs_url = "http://testserver{}".format(url)
        client = Client()

        # User request without ticket: redirect to the CAS login page.
        response = client.get(url)
        self.assertRedirects(
            response,
            "https://cas.eleves.ens.fr/login?service={}".format(quote(abs_url)),
            fetch_redirect_response=False,
        )

    @patch("authens.backends.get_cas_client")
    def test_cas_login_with_ticket(self, mock_cas_client):
        # Make `get_cas_client` return a dummy CAS client that skips ticket verification
        # and always log in a user with CAS login 'johndoe'.
        mock_cas_client.return_value = FakeCASClient()

        url = reverse("authens:login.cas")
        client = Client()

        # User request with a CAS ticket: validate the ticket.
        response = client.get("{}?ticket=dummy-ticket".format(url), follow=True)
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL)

    def test_can_gen_login_switch(self):
        response = Client().get(reverse("authens:login"))
        self.assertEqual(response.status_code, 200)

    def test_oldcas_login(self):
        url = reverse("authens:login.oldcas")
        client = Client()

        user = UserModel.objects.create_user(username="johndoe31", password="password")
        # Decoy user that may be authenticated by mistake
        UserModel.objects.create_user(username="johndoe", password="password")
        OldCASAccount.objects.create(user=user, cas_login="johndoe", entrance_year=2014)

        post_data = dict(cas_login="johndoe", password="password", entrance_year=2014)
        response = client.post(url, post_data)
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL)

    def test_oldcas_login_error(self):
        url = reverse("authens:login.oldcas")
        client = Client()

        user = UserModel.objects.create_user(username="johndoe31", password="password")
        OldCASAccount.objects.create(user=user, cas_login="johndoe", entrance_year=2014)

        wrong_year = dict(cas_login="johndoe", password="password", entrance_year=2015)
        wrong_login = dict(
            cas_login="johndoe31", password="password", entrance_year=2014
        )

        response = client.post(url, wrong_year, follow=True)
        non_field_errors = response.context["form"].non_field_errors().as_data()
        self.assertEqual(non_field_errors[0].code, "invalid_login")

        response = client.post(url, wrong_login, follow=True)
        non_field_errors = response.context["form"].non_field_errors().as_data()
        self.assertEqual(non_field_errors[0].code, "invalid_login")


class TestLogoutView(TestCase):
    def test_regular_logout(self):
        # Regular user (without a CAS account)
        user = UserModel.objects.create_user(username="johndoe")

        # Log the user in
        client = Client()
        client.force_login(user)

        self.assertEqual(Session.objects.count(), 1)
        response = client.get("/authens/logout")
        self.assertEqual(Session.objects.count(), 0)  # User is actually logged out.
        self.assertRedirects(response, settings.LOGOUT_REDIRECT_URL)

    @patch("authens.backends.get_cas_client")
    def test_cas_logout(self, mock_cas_client):
        # Make `get_cas_client` return a dummy CAS client that skips ticket verification
        # and always log in a user with CAS login 'johndoe'.
        # This is only used for login.
        mock_cas_client.return_value = FakeCASClient()

        # CAS user
        user = UserModel.objects.create_user(username="johndoe")
        CASAccount.objects.create(user=user, cas_login="johndoe", entrance_year=2019)

        # Log the user in via CAS
        client = Client()
        client.get("/authens/login/cas?ticket=dummy-ticket")

        self.assertEqual(Session.objects.count(), 1)
        response = client.get("/authens/logout")
        self.assertEqual(Session.objects.count(), 0)  # User is logged out…
        self.assertRedirects(  # … and redirected to the CAS logout page.
            response,
            "https://cas.eleves.ens.fr/logout?service={}".format(
                quote("http://testserver/public")
            ),
            fetch_redirect_response=False,
        )

    def test_regular_logout_on_cas_account(self):
        # CAS user
        user = UserModel.objects.create_user(username="johndoe", password="p4ssw0rd")
        CASAccount.objects.create(user=user, cas_login="johndoe", entrance_year=2019)

        # Log the user in by password and *not* via CAS
        client = Client()
        client.login(username="johndoe", password="p4ssw0rd")

        self.assertEqual(Session.objects.count(), 1)
        response = client.get("/authens/logout")
        self.assertEqual(Session.objects.count(), 0)  # User is logged out…
        # … and not redirected to the CAS logout page.
        self.assertRedirects(response, settings.LOGOUT_REDIRECT_URL)



from django.contrib.auth.decorators import login_required
from django.shortcuts import render


def home(request):
    ctx = {"user": request.user}
    return render(request, "example/home.html", ctx)


@login_required
def protected_page(request):
    return render(request, "example/protected.html")



# Generated by Django 3.0.6 on 2020-06-12 17:26
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("authens", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="OldCASAccount",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "cas_login",
                    models.CharField(max_length=1023, verbose_name="ancien login CAS"),
                ),
                (
                    "entrance_year",
                    models.SmallIntegerField(
                        verbose_name="année de création du compte CAS"
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="old_cas_account",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="utilisateurice",
                    ),
                ),
            ],
            options={
                "verbose_name": "Ancien compte CAS",
                "verbose_name_plural": "Anciens comptes CAS",
            },
        ),
        migrations.AddConstraint(
            model_name="oldcasaccount",
            constraint=models.UniqueConstraint(
                fields=("cas_login", "entrance_year"), name="clipper_year_uniqueness"
            ),
        ),
    ]



from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.urls import include, path

# ---
# Two tiny views to easily test user authentication.
# ---


def public_view(request):
    return HttpResponse("OK")


@login_required
def private_view(request):
    return HttpResponse("OK")


# ---
# Urls: expose authens' urls + the above views.
# ---


urlpatterns = [
    path("public", public_view),
    path("private", private_view),
    path("authens/", include("authens.urls")),
]
